#!/usr/local/bin/perl
#
# info2www - Gateway between GNU Info nodes and WWW
# $Id: info2www,v 1.1 1994/03/24 19:15:32 lmdrsm Exp $
#
# This is a script conforming to the CGI - Common Gateway Interface
#
# Created:      1994-03-24 by LMD/T/AD Roar Smith (lmdrsm@lmd.ericsson.se)
#
# Copyright:    This program is in the Public Domain.
#
# The original code (most of &info2html) was written by
# Eelco van Asperen (evas@cs.few.eur.nl).
#
# $Log: info2www,v $
# Revision 1.1  1994/03/24  19:15:32  lmdrsm
# Included automatic reference to the info2www documentation.
#
# Revision 1.0  1994/03/24  17:49:15  lmdrsm
# Initial revision
#
#
#

#----------------- CONFIGURATION ----------------------------------------------
-

#
# Set $debug = 1; to debug what's happening
#
$debug = 0;

#
# INFOPATH is the path of direcories in which to search for Info node files.
#
@INFOPATH =
    (
     "/usr/local/gnu/info",
     "/usr/local/info"
     );

#
# ALIAS is a map of aliases - look for the alias if the node itself isn't found
.
# The key (first entry) is the node filename, the value (second entry) is the
# alias. Both are basenames (i.e. no path!) with no capital letters.
# Note that they keys *must* be unique!
#
%ALIAS =
    (
     'emacs',   'lemacs',
     'g++',     'gcc',
     'c++',     'gcc',
     'gunzip',  'gzip',
     'zcat' ,   'gzip'
     );

#
# Location of the icons used for indicating references and stuff:
# $info_icon    - Icon at the top left of each document
# $up_icon      - Icon used in an "Up:"   hyperlink at the top
# $next_icon    - Icon used in a  "Next:" hyperlink at the top
# $prev_icon    - Icon used in a  "Prev:" hyperlink at the top
# $menu_icon    - Icon used in front of each menu label
#
# Set these to "" if you don't want them used.
#
$info_icon =    "http://www.ph.unimelb.edu.au/info2www/infodoc.gif";
$up_icon =      "http://www.ph.unimelb.edu.au/info2www/up.gif";
$next_icon =    "http://www.ph.unimelb.edu.au/info2www/next.gif";
$prev_icon =    "http://www.ph.unimelb.edu.au/info2www/prev.gif";
$menu_icon =    "http://www.ph.unimelb.edu.au/info2www/menu.gif";

#
# Location of documentation for info2www
#
# Set this to "" if you don't want it used.
#
$docref =       "http://www.ph.unimelb.edu.au/info2www/info2www.html";

#----------------- MAIN -------------------------------------------------------
-
print "Content-type: text/html\n";  #-- Mime header for NCSA httpd
print "\n";

require 'sys/socket.ph';

$pg = $0; $pg =~ s,^.*/([^/]*)$,$1,;
$options = "";
$script_name = $ENV{'SCRIPT_NAME'};
$server_name = $ENV{'SERVER_NAME'};
$request_method = $ENV{'REQUEST_METHOD'};
$prefix = $script_name . "?";   # prefix for HREF= entries

if ($request_method ne 'GET') {
    die "REQUEST_MODE 'GET' expected - got '$request_method'\n";
}

print "ARGV: ", join('+', @ARGV), "<BR>\n" if $debug;
if ($#ARGV == -1) {
    $nodename = "(DIR)";
} else {
    $nodename = join('+', @ARGV);
    $nodename = &DeEscape($nodename);
}
print "nodename: ", $nodename, "<BR>\n" if $debug;

&info2html($nodename);

if ($docref) {
    print
        "<hr>\n",
        "<em>automatically generated by</em> ",
        "<A HREF=\"$docref\"><strong>$pg</strong></A>\n";
} else {
    print
        "<hr>\n",
        "<em>automatically generated by</em> ",
        "<strong>$pg</strong>\n";
}

exit(0);

#----------------- SUBROUTINES ------------------------------------------------
-

#------------------------------------------------------------
#                        ToPattern
#------------------------------------------------------------
# This procedure transforms a string in a search pattern,
# escaping the non standard characters.
#------------------------------------------------------------
sub ToPattern{
  local($Tag) = @_;
  local(@Temp);
  @Temp = split(/([^a-zA-Z0-9])/,$Tag);
  $Tag = "";
  for $x (@Temp){
    $x = ($x =~ /[^a-zA-Z0-9]/) ? '\\'.$x : $x;
    $Tag .= $x;
  }
  $Tag;
}

#---------------------------------------------------------
#                      Escape
#---------------------------------------------------------
#  This procedures escapes some special characeters. The
#  escape sequence follows the WWW guide for escaped
#  characters in URLs
#---------------------------------------------------------
sub Escape{
  local($Tag) = @_;
  $Tag =~ s/ /%20/g;     #  space
  $Tag =~ s/\+/%2B/g;    #  +
  return $Tag;
}

#----------------------------------------------------------
#                    DeEscape
#----------------------------------------------------------
sub DeEscape{
  local($Tag) = @_;
  $Tag =~ s/\\([][(){}|?*\\])/$1/g;
  return $Tag;
}


#---------------------------------------------------------------------------
#
#                    info2html
#
#---------------------------------------------------------------------------
sub info2html {
    local($nodename) = @_;
    local($next_img, $prev_img, $up_img);

    # Nodename looks like one of these:
    # (file)label       - Both file and label of the Info node given
    # (file)            - Label defaults to "Top"
    #                   - File defaults to "DIR", Label defaults to "Top"

    $matches = 0;
    $blank = 0;

    if ($nodename =~ /^\(([^\)]*)\)(.+)$/) {
        ($file, $node) = ($1, $2);
    } elsif ($nodename =~ /^\(([^\)]*)\)$/) {
        ($file, $node) = ($1, "Top");
    } elsif (!$nodename) {
        ($file, $node) = ("DIR", "Top");
    } else {
        print "Malformed node name: $nodename\n";
        return(0);
    }

    $target = $node;
    $target =~ y/A-Z/a-z/;
    $target =~ s/%20/ /g;
    $target =~ s/&lt\;/</g;
    $target =~ s/&gt\;/>/g;
    $target = &ToPattern($target);
    $file =~ s/&lt\;/</g;
    $file =~ s/&gt\;/>/g;

    print "nodename: $nodename\nfile: $file\ntarget: $target\n" if $debug;

    $info_img = "<IMG SRC=\"$info_icon\" ALIGN=BOTTOM> " if $info_icon;
    $next_img = "<IMG SRC=\"$next_icon\" ALIGN=BOTTOM> " if $next_icon;
    $prev_img = "<IMG SRC=\"$prev_icon\" ALIGN=BOTTOM> " if $prev_icon;
    $up_img = "<IMG SRC=\"$up_icon\" ALIGN=BOTTOM> " if $up_icon;

    $nfiles = 0;
    &OpenFile($file) || return(0);

    $active = 0;
    $seenMenu = 0;
    $indirect = 0;
    $inentry = 0;
    $lastblank = 0;

  FileLoop:
    for (; $nfiles > 0; ) {

        local($handle) = "FH_$nfiles";
        print "--now reading from $handle--<P>\n" if $debug;
        while (<$handle>) {
            chop;
            s/</&lt\;/g;
            s/>/&gt\;/g;
            #study;                     # Doesn't seem to help or hurt!
            /^[\037\f]/ && do {
                &EndMenu();
                &EndListing();
                return(0) if $active;   # we've done our work
                $active = 0;
                $seenMenu = 0;
                $indirect = 0;
                $inentry = 0 if $inentry;
                $inentry++;
                next;
            };

            next if ($inentry == 0);

            $lastblank = $blank; $blank = 0;
            /^$/ && do {
                if ($active) {
                    print "\n";
                } elsif ($menu == 0) {
                    print;
                }
                $blank = 1;
                next;
            };

            ($inentry == 1) && do  {
                # top line:
                # File: info,  Node: Add,  Up: Top,  Previous: Expert,  Next: M
enus
                /^tag table:/i && do {
                    # we don't use the tag table
                    $inentry = 0;
                    next;
                };
                /^indirect:/i && do {
                    # this entry is a list of filenames to include:
                    #
                    #   gcc.info-1: 1131
                    #   gcc.info-2: 49880
                    #   gcc.info-3: 99426
                    $inentry++;
                    $indirect++;
                    next;
                };

                #
                # Parse the header line. If one of the fields
                #       Node: Up: Next: Previous: File:
                # is found, then a variable 'h_node' is set for
                # the field 'node:', 'h_next' for 'next:', etc.
                #
                undef $h_node;
                undef $h_file;
                undef $h_next;
                undef $h_prev;
                undef $h_up;

                @F = split(/[,\s]+/);
                for ($i = 0; $i <= $#F; ) {
                    $_ = $F[$i];
                    if ( /^node:$/i || /^file:$/i || /^up:$/i
                        || /^previous:$/i || /^prev:$/i || /^next:$/i ) {
                        #
                        # make the name of the variable:
                        #
                        $v = $_;
                        $v =~ s/://;
                        $v =~ tr/A-Z/a-z/;
                        $v = "prev" if $v eq "previous";
                        $w = "";
                        #
                        # find all words that belong to this field:
                        #
                        for ($i++; $i <= $#F; $i++) {
                            $_ = $F[$i];
                            last if ( /^node:$/i || /^file:$/i
                                     || /^up:$/i || /^previous:$/i
                                     || /^prev:$/i || /^next:$/i );
                            $w .= $_ . " ";
                        }
                        $w =~ s/\s+$//;         # delete trailing spaces
                        eval "\$h_$v = \$w;";
                    } else {
                        $i++;
                    }
                }
                
                print "--h_node: $h_node--<p>\n" if $debug;
                $n = 0;

                if ($h_node =~ m/^$target/i) {
                    $active = 1;
                    $matches++;
                    print
                        "<TITLE>",
                        "Info Node: ($file)$h_node",
                        "</TITLE>\n",
                        "<H1>$info_img($file)$h_node</H1>\n",
                        "<HR>\n";
                    if (defined $h_next) {
                        print
                            "Next: ",
                            "<B>",
                            &make_anchor($h_next, "$next_img$h_next"),
                            "</B><TT>  </TT>";
                        $n++;
                    }
                    if (defined $h_prev) {
                        print
                            "Prev: ",
                            "<B>",
                            &make_anchor($h_prev, "$prev_img$h_prev"),
                            "</B><TT>  </TT>";
                        $n++;
                    }
                    if (defined $h_up) {
                        print
                            "Up: ",
                            "<B>", &make_anchor($h_up, "$up_img$h_up"),
                            "</B><TT>  </TT>";
                        $n++;
                    }
                }

                print "\n<HR>\n" if $n;
                $inentry++;
                &StartListing();
                next;
            };

            ($inentry == 2) && $indirect && do  {
                # each line of this entry consists of two fields,
                # a filename and an offset, separated by a colon.
                # For example:
                #       texinfo-1: 1077
                local(@F) = split(/:/);
                print "#include $F[0]<p>\n" if $debug;
                # should save: $inentry $indirect
                $save_inentry[$nfiles] = $inentry;
                $save_indirect[$nfiles] = $indirect;
                $inentry = 0;
                $indirect = 0;
                &OpenFile($F[0]) || return(0);
                next FileLoop;
            };

            next if $active == 0;

            if (($end) = /^\*\s+Menu:(.*)$/) {
                # start of a menu:
                $seenMenu = 1;
                &EndListing();
                print "$end";
                &StartMenu();
                next;
            };

            /^\*/ && do {
                #---- SAMPLE LINES: -----------------------------------------
                # * Sample::.           Sample info.
                #
                # * Info: (info).       Documentation browsing system.
                #
                # * Bison: (bison/bison)
                #               A Parser generator in the same style as yacc.
                # * Random: (Random) Random    Random Number Generator
                #------------------------------------------------------------

                if ($menu == 0 && $seenMenu) { &EndListing(); &StartMenu(); };

                # * foo::
                /^\*\s+([^:]+)::/ && do {
                    $rest_of_line = $';
                    print
                        "<DT>", &make_anchor($1, $1, $menu_icon),
                        "<DD>";
                    $rest_of_line =~ s/^[\s\.]+//;
                    print $rest_of_line, "\n";
                    next;
                };

                # * foo: (bar)beer OR (bar)
                /^\*\s+([^:]+):\s+\(([^\) \t\n]+)\)([^\t\n\.,]*)/ && do {
                    $rest_of_line = $';
                    print
                        "<DT>", &make_anchor("($2)$3",$1, $menu_icon),
                        "<DD>";
                    $rest_of_line =~ s/^[\s\.]+//;
                    print $rest_of_line, "\n";
                    next;
                };

                # * foo: beer.
                /^\*\s+([^:]+):\s+([^\t,\n\.]+)/ && do {
                    $rest_of_line = $';
                    print
                        "<DT>", &make_anchor($2, $1, $menu_icon),
                        "<DD>", $2;
                    $rest_of_line =~ s/^[\s\.]+//;
                    print $rest_of_line, "\n";
                    next;
                };

                # no match: ignore silently
            };

            $menu && $lastblank && do {
                &EndMenu();
                &StartListing();
            };

            $menu && do {
                s/^\s+//;
            };

            /\*note/i && do {
                # cross reference entry:
                # "*note nodename::."
                # "*note Cross-reference-name: nodename."
                local($n) = 0;
                while (1) {
                    # *note foo\nbar... (reference split over newline)
                    if (/\*note\s+([^:\.]+)$/i) {
                        $_ .= "\n" . <$handle>; # Merge with next line
                        chop;
                    }

                    # *note foo:
                    if (/\*note\s+([^:\.]+)::/i) {
                        s//@@@NOTE@@@/; # insert unique (I hope) marker
                        local($ref) = sprintf("<B>Note:</B> %s",
                                              &make_anchor($1, $1));
                        s/@@@NOTE@@@/$ref$1<\/A>/;
                        $n++;
                        next;
                    }

                    # * foo: (bar)beer OR (bar)
                    if (/\*note\s+([^:]+):\s+\(([^\) \t\n]+)\)([^\t\n\.,]*)/i)
{
                        s//@@@NOTE@@@/; # insert unique (I hope) marker
                        local($ref) = sprintf("<B>Note:</B> %s",
                                              &make_anchor("($2)$3", $1));
                        s/@@@NOTE@@@/$ref/;
                        $n++;
                        next;
                    }

                    # * foo: beer.
                    if (/\*note\s+([^:]+):\s+([^\t,\n\.]+)/i) {
                        s//@@@NOTE@@@/; # insert unique (I hope) marker
                        local($ref) = sprintf("<B>Note:</B> %s",
                                              &make_anchor($2, $1));
                        s/@@@NOTE@@@/$ref/;
                        $n++;
                        next;
                    }

                    last;
                }
#               if ($n > 0) {
#                   local($l) = $listing;
#                   &EndListing() if $l;
#                   print "$_\n";
#                   &StartListing() if $l;
#                   next;
#               }
            };

            print "$_\n";
        }
        &EndMenu();

        # clear status variables;
        $active = 0;
        $seenMenu = 0;
        $indirect = 0;
        $inentry = 0;
        $lastblank = 0;

        print "--end of file $handle--<P>\n" if $debug;
        $nfiles--;
        $inentry = $save_inentry[$nfiles];
        $indirect = $save_indirect[$nfiles];
        print "--inentry: $inentry--indirect: $indirect--<p>\n" if $debug;
    }
    if ($matches == 0) {
        print "Sorry, entry '$target' was not found.\n";
    }
}

#---------------------------------------------------------------------------

sub make_anchor {
    local($ref, $label, $icon) = @_;
    local($node_file, $node_name, $img, $href);

    print "--make_anchor($ref, $label)<BR>\n" if $debug;
    # (foo)bar
    if ($ref =~ m/\(([^\)]+)\)\s*([^\t\n,\.]*)/) {
        $node_file = $1;
        $node_name = $2;
    } elsif ($file =~ /^dir$/i) {
        print "--(DIR) node - Menu \"@_\" means \"($ref)\"<BR>\n" if $debug;
        $node_file = $ref;
        $node_name = "";
    } else {
        $node_file = $file;
        $node_name = $ref;
    }
    $node_name =~ s/[ ]*$//;
    $node_name =~ s/ /%20/g;

    if ($node_name ne "") {
        $href = &Escape("$prefix($node_file)$node_name");
    } else {
        $href = &Escape("$prefix($node_file)");
    }
    if ($icon) {
        $img = "<IMG SRC=\"$icon\"> ";
    }
    return "$img<A HREF=\"$href\">$label</A>";
}

sub StartMenu {
    print "\n<DL>" if $active;
    $menu = 1;
}

sub EndMenu {
    if ($menu) {
        print "</DL>\n" if $active;
        $menu = 0;
    }
}

sub StartListing {
    print "<PRE>\n" if $active;
    $listing++;
}

sub EndListing {
    if ($listing) {
        print "</PRE>\n" if $active;
        $listing--;
    }
}

sub OpenFile {
    local($filename) = @_;
    local($alternate, $handle, $success);

    $filename =~ y/A-Z/a-z/;
    $alias = $ALIAS{$filename};

    $nfiles++;
    $handle = "FH_$nfiles";
    $success = 0;
    print "<P>Trying to open file",
    "\"$filename\" or \"$filename.info\" in INFOPATH...\n" if $debug;
    foreach (@INFOPATH) {
        if (open($handle, "$_/$filename")) {
            $success = 1;
            print "<P>Opened file \"$_/$filename\"\n" if $debug;
            last;
        } elsif (open($handle, "$_/$filename.info")) {
            $success = 1;
            print "<P>Opened file \"$_/$filename.info\"\n" if $debug;
            last;
        }
    }
    if (!$success && $alias) {
        print "<P>Trying to open alias file",
        "\"$alias\" or \"$alias.info\" in INFOPATH...\n" if $debug;
        foreach (@INFOPATH) {
            if (open($handle, "$_/$alias")) {
                $success = 1;
                print "<P>Opened file \"$_/$alias\"\n" if $debug;
                last;
            } elsif (open($handle, "$_/$alias.info")) {
                $success = 1;
                print "<P>Opened file \"$_/$alias.info\"\n" if $debug;
                last;
            }
        }
    }
    if (!$success) {
        print "<P>Could not open file",
        "\"$filename\" or \"$filename.info\" in INFOPATH.\n";
        print "<P>INFOPATH = (<BR>\n", join("<BR>\n", @INFOPATH), "<BR>)\n" if
$debug;
        return(0);
    }
    return(1);
}

sub error {
    local($reason) = @_;

    print
"<TITLE>Lookup Error</TITLE>
<H1>Lookup Error</H1>
Can't retrieve your request - $reason\n";

    exit(0);
}

#---------------------------------------------------------------------------

